# -*- coding: utf-8 -*-
"""
-------------------------------------------------
   File Name：     graphSvg
   Description :
   Author :       liaozhaoyan
   date：          2022/10/4
-------------------------------------------------
   Change Activity:
                   2022/10/4:
-------------------------------------------------
"""
__author__ = 'liaozhaoyan'

from treelib import Tree


class CgraphTree(object):
    def __init__(self):
        super(CgraphTree, self).__init__()

    def _splitLine(self, line):
        rd = {}
        cpu, res = line.split(")", 1)
        rd['cpu'] = int(cpu)

        us, symbol = res.split("|", 1)
        us = us.strip()
        if us.endswith("us"):
            if us[0] in ('+', '!', '#', '*'):
                rd['us'] = float(us[1:-2])
            else:
                rd['us'] = float(us[:-2])
        else:
            rd['us'] = None

        rd['symbol'] = symbol.strip()
        return rd

    def tree(self, lines):
        stack = []
        rtree = Tree()
        last = rtree.root
        for line in lines:
            lined = self._splitLine(line)
            us = lined['us']
            symbol = lined['symbol']
            if us is None:
                if '(' in symbol:
                    func, _ = symbol.split('(', 1)
                    node = rtree.create_node(tag=func, parent=last, data={"func": func})
                    stack.append(node)
                    last = node
            else:
                if symbol.endswith(";"):
                    func, _ = symbol.split('(', 1)
                    rtree.create_node(tag=func, parent=last, data={"func": func, "us": us})
                else:
                    last.data['us'] = us
                    stack.pop()
                    if len(stack) > 0:
                        last = stack[-1]
        root = rtree.root
        if 'us' not in rtree[root].data:
            raise IndexError("bad index.")
        return rtree

    def _filterChildren(self, mtree, nid, tag):
        for child in mtree.children(nid):
            if child.tag == tag:
                return child
        return None

    def mergeTrees(self, trees):
        summary = Tree()
        stack = []
        for tree in trees:
            for nid in tree.expand_tree(mode=Tree.DEPTH, sorting=False):
                node = tree.get_node(nid)
                level = tree.level(nid)
                if level == 0:  # root level
                    if len(stack) == 0:
                        last = summary.root
                        mNode = summary.create_node(tag=node.tag, parent=last, data=node.data)
                        stack.append(mNode)
                    else:
                        mNode = stack[0]
                        mNode.data['us'] += node.data['us']
                        stack = stack[:1]
                else:  # next level
                    last = stack[level - 1]
                    mNode = self._filterChildren(summary, last.identifier, node.tag)
                    if mNode:
                        mNode.data['us'] += node.data['us']
                    else:
                        mNode = summary.create_node(tag=node.tag, parent=last, data=node.data)
                    stack = stack[:level]
                    stack.append(mNode)
        return summary

    def walk(self, tree):
        for nid in tree.expand_tree(mode=Tree.DEPTH, sorting=False):
            node = tree.get_node(nid)
            level = tree.level(nid)
            print(level, node.data)


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
"""
-------------------------------------------------
   File Name：     sobuild
   Description :
   Author :       liaozhaoyan
   date：          2022/7/1
-------------------------------------------------
   Change Activity:
                   2022/7/1:
-------------------------------------------------
"""
__author__ = 'liaozhaoyan'

import os
import sys
from .kobuild import CkoBuilder


class Csobuid(CkoBuilder):
    def __init__(self):
        super(Csobuid, self).__init__()

    def build(self, iPath, out="user.db"):
        pass


def main():
    k = CkoBuilder()
    k.build(sys.argv[1])


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
"""
-------------------------------------------------
   File Name：     surfGraph
   Description :
   Author :       liaozhaoyan
   date：          2022/10/4
-------------------------------------------------
   Change Activity:
                   2022/10/4:
-------------------------------------------------
"""
__author__ = 'liaozhaoyan'

import sys
import argparse
import time
import re
import pickle
from collections import deque
from treelib import Tree
from .ftrace import ftrace
from .graphTree import CgraphTree
from quickSvg import Flame


def getValue(data):
    return data['us']


def getNote(tree, node):
    root = tree.get_node(tree.root)
    perRoot = node.data['us'] / root.data['us'] * 100.0

    parent = tree.parent(node.identifier)
    if parent is None:
        perParent = 100.0
    else:
        perParent = node.data['us'] / parent.data['us'] * 100.0
    return "cost %f us, %f%% from root, %f%% from parent" % (node.data['us'], perRoot, perParent)


class surfGraph(ftrace):
    def __init__(self, symbol, mode="raw", output=True, step=True):
        super(surfGraph, self).__init__(show=False, echo=True, instance=None)
        self._symbol = symbol
        self._record = 'nop'

        cbD = {
            "raw": self._cbLine,
            'tree': self._cbTree,
            "walk": self._cbWalk,
            "svg": self._cbSvg,
        }
        if mode not in cbD.keys():
            raise ValueError("mode only support %s" % str(cbD.keys()))
        self._cb = cbD[mode]
        self._funStart = False
        self._reStart = re.compile(r"^[\d]+\) +\|  [\w]+\(\) \{")
        self._reEnd = re.compile(r"^[\d]\).+\|  \}")
        self._reLine = re.compile(r"^[\d]\).+\|")

        self._serial = 1
        self._lines = []
        self._output = output
        self._step = step
        self._tree = None
        self._trees = deque(maxlen=32)

    def _setupTracer(self, symbol):
        Path = self.baseDir + "/tracing/current_tracer"
        self._record = self._c.cmd("cat %s" % Path)
        self._echoPath(Path, "nop")

        time.sleep(0.5)
        fPath = self.baseDir + "/tracing/set_graph_function"
        self._echoPath(fPath, symbol)
        self._echoPath(Path, "function_graph")

    def _saveTree(self, tree):
        self._trees.append(tree)
        if self._tree is None:
            self._tree = tree
        else:
            t = CgraphTree()
            self._tree = t.mergeTrees([self._tree, tree])

    def _setupTree(self, line, cbTree):
        if self._funStart:
            if self._reLine.search(line):
                self._lines.append(line)
                if self._reEnd.search(line):
                    self._funStart = False
                    t = CgraphTree()
                    try:
                        tree = t.tree(self._lines)
                    except IndexError:
                        return

                    node = tree.get_node(tree.root)
                    if 'us' not in node.data:
                        return

                    if self._output:
                        self._saveTree(tree)
                    cbTree(tree)
        elif self._reStart.search(line):
            res = line.split("|")[1].strip()
            res = res.split("(")[0]
            if res == self._symbol:
                self._funStart = True
                self._lines = [line]

    def _cbLine(self, line):
        print("%s" % line)

    def _treeShow(self, tree):
        print(tree)

    def _walkShow(self, tree):
        for nid in tree.expand_tree(mode=Tree.DEPTH, sorting=True):
            node = tree.get_node(nid)
            level = tree.level(nid)
            print(level, node.data)

    def _svgShow(self, tree):
        if self._step:
            print("save %s-%d.svg" % (self._symbol, self._serial))
            svg = Flame("%s-%d.svg" % (self._symbol, self._serial))
            svg.render(tree, getValue, getNote, "%s" % self._symbol)
        else:
            print("jump %s-%d.svg" % (self._symbol, self._serial))
        self._serial += 1

    def _cbTree(self, line):
        self._setupTree(line, self._treeShow)

    def _cbWalk(self, line):
        self._setupTree(line, self._walkShow)

    def _cbSvg(self, line):
        self._setupTree(line, self._svgShow)

    def procLine(self, line):
        # print(line)
        self._cb(line)

    def start(self):
        func = self._checkAvailable(self._symbol)
        if func is None:
            raise ValueError("symbol %s is not an available function." % self._symbol)

        self._setupTracer(func)
        super(surfGraph, self).start()

    def stop(self):
        super(surfGraph, self).stop()
        Path = self.baseDir + "/tracing/set_graph_function"
        self._echoPath(Path, '')

        if self._tree:
            if self._output:
                dOut = {"merge": self._tree,
                        "trees": self._trees,
                        }
                s = pickle.dumps(dOut, protocol=2)
                with open("%s.tree" % self._symbol, 'wb') as f:
                    f.write(s)

            if self._cb == self._cbSvg:
                svg = Flame("%s.svg" % self._symbol)
                svg.render(self._tree, getValue, getNote, "%s" % self._symbol)
                print("write %s.svg" % self._symbol)


def main():
    examples = """examples: surfGraph -f __do_fault"""

    parser = argparse.ArgumentParser(
        description="kernel function call graph tool.",
        formatter_class=argparse.RawDescriptionHelpFormatter,
        epilog=examples
    )

    parser.add_argument('-f', '--function', type=str, dest='function', default="", help='set function to call graph.')
    parser.add_argument('-m', '--mode', type=str, dest='mode', default="svg",
                        help="set output mode, support svg(default)/tree/walk/raw")
    parser.add_argument('-s', '--step', type=int, dest='step', default=True,
                        help="write file by every step, only for svg mode.")
    parser.add_argument('-o', '--output', type=int, dest='output', default=True,
                        help="save trees to *.tree file, 32 max")

    args = parser.parse_args()
    symbol = args.function
    graph = surfGraph(symbol, mode=args.mode, output=args.output, step=args.step)
    graph.start()
    graph.loop()


if __name__ == "__main__":
    pass



# -*- coding: utf-8 -*-
"""
-------------------------------------------------
   File Name：     uprobeParser
   Description :
   Author :       liaozhaoyan
   date：          2022/9/23
-------------------------------------------------
   Change Activity:
                   2022/9/23:
-------------------------------------------------
"""
__author__ = 'liaozhaoyan'

import re
import os
from .execCmd import CexecCmd
from .gdbParser import CgdbParser
from .surfException import FileNotExistException, InvalidArgsException


class CuprobeParser(CgdbParser):
    def __init__(self, obj, gdb=None):
        self._libD = {}
        obj = self._setupObj(obj)
        self._obj = obj
        super(CuprobeParser, self).__init__(obj, gdb)
        self._reBraces = re.compile(r"(?<=\{).+?(?=\})")
        self._beg = self._setupBeg()

    def _setupBeg(self):
        cmd = "objdump -x %s" % self._obj
        lines = self._cmd.cmd(cmd)
        for line in lines.split('\n'):
            s = line.strip()
            if s.startswith("LOAD off"):
                # LOAD off    0x0000000000000000 vaddr 0x0000000000000000 paddr 0x0000000000000000
                s = re.sub(" +", " ", s)
                vs = s.split(" ")
                v = vs[4]
                return int(v, 16)
        raise ValueError("file has not LOAD off segments.")

    def _setupObj(self, obj):
        if '/' in obj:
            if os.path.exists(obj):
                res = os.path.abspath(obj)
            else:
                raise InvalidArgsException("can not find file %s." % obj)
        else:
            cmd = CexecCmd()
            res = cmd.cmd("which %s" % obj)
            if res == "":
                try:
                    res = self._findLib(obj)
                    print(res)
                except KeyError:
                    raise InvalidArgsException("can not find lib %s" % obj)
        return res

    def _loadLib(self):
        if len(self._libD) == 0:
            cmd = CexecCmd()
            lines = cmd.cmd("ldconfig -v")
            path = ""
            for line in lines.split('\n'):
                if line.startswith("/"):
                    path = line.split(":", 1)[0]
                else:
                    lib, so = line.split('->', 1)
                    lib = lib.strip()
                    if ".so" in lib:
                        lib = lib.split(".so", 1)[0]
                    self._libD[lib] = os.path.join(path, so.strip())

    def _findLib(self, obj):
        self._loadLib()
        return self._libD[obj]

    def fullObj(self):
        return self._obj

    def funAddr(self, func):
        self._write("p %s" % func)
        # $3 = {<text variable, no debug info>} 0x48a870 <readline>
        try:
            res = self._read().split('\n')[-2]
        except IndexError:
            raise InvalidArgsException("can not find function %s in file %s" % (func, self._obj))
        addr = res.split(" ")[-2]
        v = int(addr, 16) - self._beg
        return "0x%x" % v

    def getFunc(self, func, ret=None, arg=None):
        self._setupRes()
        if len(func) < 2:
            return {"log": "func len should bigger than 2.", "res": None}

        File = ""
        lineNo = '0'
        self._res['res'] = []
        self._write("p %s" % func)
        try:
            res = self._read().split('\n')[-2]
        except IndexError:
            raise InvalidArgsException("can not find function %s in file %s" % (func, self._obj))

        sFinds = self._reBraces.findall(res)
        if len(sFinds):
            sFind = sFinds[0]
            if sFind == "<text variable, no debug info>":  # no debug symbol.
                ret, args = "int", "int, int, int, int"
            else:
                ret, args = sFind.split('(', 1)  # int (void)
                ret = ret.strip()
                args = args[:-1]
            funcd = {'func': func,
                     'args': self._argFuncSplit(args),
                     'ret': ret,
                     'line': int(lineNo),
                     'file': File}
            self._res['res'].append(funcd)
        return self._res


if __name__ == "__main__":
    pass
